{-# OPTIONS_GHC 
         -XRank2Types -XTypeOperators
         -XFlexibleInstances -XFlexibleContexts
         -XScopedTypeVariables 
#-}

module MOS where

import Prelude hiding (seq) 
import Control.Monad
import Interpreters
import AutoLift hiding (Put)

----------------------------------------------------------------
-- Transition Relations as Coalgebras 
----------------------------------------------------------------

data BP a = BP { unBP :: NonDet (Either () (Pr a)) }

instance Functor BP where
   fmap f (BP m) = BP (fmap (fmap (fmap f)) m)

newtype KI a    = KI Int

instance Functor KI where
    fmap _ (KI c) = KI c
data KE a    = KE

instance Functor KE where
    fmap _ KE = KE

data Nu f = Nu (f (Nu f)) 

out         :: (Nu f) -> f (Nu f)
out (Nu n)  = n

unfold    :: Functor b => (x -> b x) -> x -> Nu b
unfold g  = Nu . fmap (unfold g) . g

----------------------------------------------------------------
-- Mathematical Operational Semantics 
----------------------------------------------------------------

type OR s b =  forall x. s (x, b x) -> b (Term s x) 

orP                      :: OR P BP
orP Nil                  = BP (return (Left ()))
orP (Put c)              = BP (return (Right (Pr c nil)))
orP (Seq (_,bp) (q,bq))  = BP (unBP  bp >>= either  
                                (\() -> unBP (fmap Var bq))
                                (\(Pr c p')  -> return (Right (Pr c (seq (Var p') (Var q))))))
orP (Alt (_,bp) (_,bq))  = BP (unBP (fmap Var bp) `plusND` unBP (fmap Var bq))

----------------------------------------------------------------
-- Obtaining a transition relation 
----------------------------------------------------------------

opMonad       ::  (Functor s, Functor b) =>
                  OR s b -> (x -> b x) -> Term s x -> b (Term s x)
opMonad op k  = snd . foldTerm  (pair Var (fmap Var . k)) 
                                 (pair (Con . fmap fst) (fmap join . op))
                  where  pair f g a = (f a, g a)

run         ::  (Functor s, Functor b) => OR s b -> Program s -> Nu b
run op      = unfold (opMonad op empty)

denModel     :: (Functor s, Functor b) => OR s b -> s (Nu b) -> Nu b
denModel or  = unfold (opMonad or out) . Con . fmap Var

eval         :: (Functor s, Functor b) => OR s b -> Program s -> Nu b
eval or      = foldTerm empty (denModel or)

type B m f v x = m (Either v (f x))

----------------------------------------------------------------
-- Modular Operational Semantics
----------------------------------------------------------------

type MOR s t m f v  =  forall x . s (x, B m f v x) -> B m f v (Term t x)

newtype BF m f v y = BF {unBF :: m (Either v (f y)) }

instance (Monad m, Functor f) => Functor (BF m f v) where
    fmap f (BF m) = BF (liftM (fmap (fmap f)) m)

ossify      :: (Functor s) => MOR s s m f v -> OR s (BF m f v)
ossify mor  = BF . mor . fmap (\(a,b) -> (a,unBF b))

----------------------------------------------------------------
-- Joining MOR
----------------------------------------------------------------

infixr 5 <+>

----------------------------------------------------------------
-- Defining Modular Operational Rules 
----------------------------------------------------------------

caseAny        ::  (Monad m, Functor f) =>   
                   B m f v x -> (Term s x -> Term s x) ->
                   (v -> B m f v (Term s x)) -> B m f v (Term s x)
caseAny b t h  =  b >>= either h (return . Right . fmap (t . Var))

caseVal        ::  (Monad m, Functor f, SubTypeV u v) =>  
                   B m f v x -> (Term s x -> Term s x) ->
                   (u -> B m f v (Term s x)) -> B m f v (Term s x)
caseVal b t h  = caseAny b t (\v -> case  (prjV v) of
                                          Nothing  -> return (Left v)
                                          Just u   -> h u) 
                                 
val  :: (SubTypeV u v, Monad m) => u -> B m f v x
val  = return . Left . injV

stp  :: (SubType g f, Monad m) => g x -> B m f v x
stp  = return . Right . inj

up  :: (Monad m, Functor s, Functor f) => B m f v x -> B m f v (Term s x)
up  = liftM (fmap (fmap Var))

seqr          :: (SubType P s) => Term s x -> Term s x -> Term s x
seqr = seq

morP ::  (Functor f, Functor s, SubType P s, SubTypeV (()) v, SubType Pr f, MonadAlgModel NonDetOp m) 
                          => MOR P s m f v
morP Nil                  = val ()
morP (Put c)              = stp (Pr c nil)
morP (Seq (_,bp) (q,bq))  = caseAny bp (`seqr` (Var q)) (\_ -> up bq)
morP (Alt (_,bp) (_,bq))  = up (bp `plusND` bq)

morZ ::(Functor f, Functor s, Monad m, SubType Z s, SubTypeV Int v) => MOR Z s m f v
morZ (Num i)                     = val i
morZ (Add (p,bp) (q,bq))         = caseVal bp (`addr` Var q)
                                              (\i -> caseVal  bq (Var p `addl`)
                                                              (\j -> val (i + j :: Int)))
morZ (Ifz (_,bc) (t,bt) (e,be))  = caseVal  bc (`ifzr` (Var t,Var e))
                                            (\i -> if i==(0::Int) then up bt else up be)

addl, addr          :: (SubType Z s) => Term s x -> Term s x -> Term s x
addl = add
addr = add

ifzr          :: (SubType Z s) => Term s x -> (Term s x, Term s x) -> Term s x
ifzr = ifz

catr          :: (SubType E s) => Term s x -> Term s x -> Term s x
catr = cat
morE                    ::   (Functor s, Functor f, SubType E s, MonadAlgModel (ThrowOp ()) m,
                             MonadModel (HandleOp ()) m) => MOR E s m f v
morE Thr                  = throw ()
morE (Cat (_,bp) (q,bq))  = handle  (bp >>= return . fmap (fmap g)) (\() -> up bq)
                                    where g p' = Var p' `catr` Var q


morR :: (Functor s, Functor f, SubType R s, MonadAlgModel NonDetOp m) => MOR R s m f v 
morR (Par (p,bp) (q,bq))  =  caseAny bp (`par` Var q) (\_ -> up bq) 
                             `plusND` 
                             liftM (fmap (fmap (\y -> Var p `par` Var y))) bq

(<+>)        ::  MOR s t m f v -> MOR s' t m f v -> MOR (s :+: s') t m f v
op1 <+> op2  = copair op1 op2

orL  :: OR (P :+: Z :+: E) (BF (ExcT () NonDet) Pr (Either () Int))
orL  = ossify (morP <+> morZ <+> morE)

orR  :: OR (R :+: P :+: Z :+: E) (BF (ExcT () NonDet) Pr (Either () Int))
orR  = ossify (morR <+> morP <+> morZ <+> morE)

ep  :: OR (E :+: P) (BF (ExcT () NonDet) Pr ())
ep  = ossify (morE <+> morP)

ez  :: OR (E :+: Z) (BF (Exception ()) Id Int)
ez  = ossify (morE <+> morZ)

-------------------------------------------------------------------------
-- Showing fixpoints
-------------------------------------------------------------------------

instance (PreservesString f, PreservesString g) => PreservesString (FComp f g) where
    preservesString (Comp x)  = preservesString (fmap preservesString x)

instance (PreservesString f) => Show (Nu f) where
   show (Nu x)  = preservesString (fmap show x)

instance Show Zero where

-------------------------------------------------------------------------
-- Showing Behaviour and Syntax
-------------------------------------------------------------------------

instance PreservesString [] where
    preservesString = ("["++) . h
        where h []        = "]"
              h [x]       = x++"]"
              h (x:y:xs)  = x++", "++h (y:xs)

instance (Monad m, PreservesString m, PreservesString f, Display v) => PreservesString (BF m f v) where
    preservesString (BF m) = preservesString (fmap (either display preservesString) m)

