{-# OPTIONS_GHC -XRank2Types #-}

module Monatron ( 
  FComp(..), Cod, liftModel, liftAlgModel,
  liftExtModel, MonadT(..), FMonadT(..), MMonadT(..), Model,
  AlgModel, ExtModel, Id(..), StateT, ReaderT, ContT, ExcT(..), WriterT,
  StepT(..), caseStep,
  NonDetT, State, Reader, Writer, Exception, Cont, NonDet, 
  runState, runReader, runCont, runException, runWriter, runNonDet, runId,
  runStateT, runReaderT, runContT, runExcT, runWriterT, runNonDetT, foldNonDetT,
  collectNonDetT, modelStateT, modelReaderT, modelWriterT, modelContT, modelStepT,
  modelThrowExcT, modelHandleExcT, modelNonDetT, getX, putX, askX, stepX,
  inEnvX, abortX, callCCX, callccX, throwX, handleX, traceX, zeroNDX, plusNDX,
  StateOp(..),ReaderOp(..),ContOp(..),ThrowOp(..),HandleOp(..),WriterOp(..),
  StepOp(..), NonDetOp(..)
  ) where

import Control.Monad 
import Data.Monoid

----------------------------
-- Monad transformers
----------------------------
class MonadT t where
    treturn  :: Monad m => a -> t m a
    treturn  = lift . return 
    tbind    :: Monad m => t m a -> (a -> t m b) -> t m b
    lift     :: Monad m => m a -> t m a

class MonadT t => FMonadT t where
    tmap :: (Functor m, Functor n) => (forall b. m b -> n b) -> t m a -> t n a

class FMonadT t => MMonadT t where
    flift      :: Functor f => f a -> t f a --should coincide with lift!
    monoidalT  :: (Functor f, Functor g) => t f (t g a) -> t (FComp f g) a 
  
-------------------------------
-- Codensity Monad Transformer
-------------------------------
      
newtype Cod f a = Cod { unCod :: forall b. (a -> f b) -> f b}

down    :: Monad m => Cod m a -> m a
down c  = unCod c return 

instance MonadT Cod where
    tbind c f  = Cod (\k -> unCod c (\a -> unCod (f a) k))
    lift m     = Cod (m >>=)

instance Monad m => Monad (Cod m) where
    return  = treturn
    (>>=)   = tbind

instance Functor f => Functor (Cod f) where
    fmap f (Cod c) = Cod $ \k -> c (k . f)
      
------------------------------------------------
-- State Monad Transformer
------------------------------------------------
      
newtype StateT s m a = S { unS :: s -> m (a,s) }

runStateT      :: s -> StateT s m a -> m (a,s) 
runStateT s m  = unS m s

instance MonadT (StateT s) where
    tbind m k  = S (\s -> unS m s >>= \(a, s') -> unS (k a) s')
    lift  m    = S (\s -> m >>= \a -> return (a,s))

instance FMonadT (StateT s) where
    tmap f (S m) = S (f . m)

instance MMonadT (StateT s) where
    flift t          = S (\s -> fmap (\a -> (a,s)) t)
    monoidalT (S t)  = S (\s -> Comp $ fmap (\(S t',s') -> t' s') (t s))

instance Monad m => Monad (StateT s m) where
    return  = treturn
    (>>=)   = tbind

instance Functor m => Functor (StateT s m) where 
    fmap f (S g) = S (\s -> fmap (\(a,s') -> (f a, s')) (g s))
      
------------------------------------------------
-- Reader Monad Transformer
------------------------------------------------
      
newtype ReaderT s m a = R { unR :: s -> m a }

runReaderT      :: s -> ReaderT s m a -> m a
runReaderT s m  = unR m s

instance MonadT (ReaderT s) where
    tbind m k  = R (\s -> unR m s >>= \a -> unR (k a) s)
    lift  m    = R (\_ -> m)

instance FMonadT (ReaderT s) where
    tmap f (R m) = R (f . m)

instance Monad m => Monad (ReaderT s m) where
    return  = treturn
    (>>=)   = tbind

instance MMonadT (ReaderT s) where
    flift t          = R (\_ -> t)
    monoidalT (R t)  = R (\s -> Comp $ fmap (($ s) . unR) (t s))

instance Functor m => Functor (ReaderT s m) where
    fmap f (R g) = R (\s -> fmap f (g s))
      
------------------------------------------------
-- Exception Monad Transformer
------------------------------------------------
      
newtype ExcT x m a = X {unX :: m (Either x a)}

runExcT :: ExcT x m a -> m (Either x a)
runExcT = unX

instance MonadT (ExcT x) where  
    tbind (X m) f  = X (do  a <- m
                            case a of  Left x  -> return (Left x)
                                       Right b -> unX (f b))
    lift m         = X (liftM Right m)
    
instance FMonadT (ExcT x) where
    tmap f  = X . f . unX

instance Monad m => Monad (ExcT x m) where
    return  = treturn
    (>>=)   = tbind

instance Functor m => Functor (ExcT x m) where 
    fmap f (X m) = X (fmap (either Left (Right .f)) m)
      
------------------------------------------------
-- Writer Monad Transformer
------------------------------------------------
      
newtype WriterT w m a = W {unW :: m (a, w)}

runWriterT  :: WriterT w m a -> m (a,w)
runWriterT  = unW

instance Monoid w => MonadT (WriterT w) where  
    tbind (W m) f  = W (do  (a,w) <- m
                            (a',w') <- unW (f a)
                            return (a',w `mappend` w'))
    lift m         = W (liftM (\a -> (a,mempty)) m)
    
instance Monoid w => FMonadT (WriterT w) where
    tmap f  = W . f . unW

instance Monoid w => MMonadT (WriterT w) where
    flift t          = W (fmap (\a -> (a,mempty)) t)
    monoidalT (W t)  = W $ Comp $  fmap (\(W t',w) -> 
                                   fmap (\(a,w') -> (a,w `mappend` w')) t') $ t

instance (Monad m, Monoid w) => Monad (WriterT w m) where
    return  = treturn
    (>>=)   = tbind

instance Functor m => Functor (WriterT x m) where 
    fmap f (W m) = W (fmap (\(a,w) -> (f a, w)) m)
      
------------------------------------------------
-- Continuation Monad Transformer
------------------------------------------------
      
newtype ContT r m a = C { unC :: (a -> m r) -> m r}

runContT      :: Monad m =>  (a -> m r) -> ContT r m a -> m r
runContT k m  = unC m k

instance MonadT (ContT r) where
    tbind c f  = C (\k -> unC c (\a -> unC (f a) k))
    lift m     = C (m >>=)

instance Monad m => Monad (ContT r m) where
    return  = treturn
    (>>=)   = tbind
      
------------------------------------------------
-- Step Monad Transformer
------------------------------------------------
      
newtype StepT f m x = T {runT :: m (Either x (f (StepT f m x)))}

instance (Functor f, Monad m) => Monad (StepT f m) where
    return  = treturn
    (>>=)   = tbind

instance (Functor f, Monad m) => Functor (StepT f m) where fmap = liftM

caseStep            ::  (Functor f, Monad m) =>  
                        (a -> StepT f m x) -> (f (StepT f m a) -> StepT f m x)
                        -> StepT f m a -> StepT f m x
caseStep v c (T m)  = T (m >>= either (runT . v) (runT . c))

unfoldStepT      :: (Functor f, Monad m) => (y -> m (Either x (f y))) -> y -> StepT f m x
unfoldStepT k y  = T (liftM (fmap (fmap (unfoldStepT k))) (k y))

instance (Functor f) => MonadT (StepT f) where
    tbind t f  = caseStep f (T . return . Right . fmap (`tbind` f)) t
    lift       = T . liftM Left

instance (Functor f) => FMonadT (StepT f) where
    tmap t (T m) = T (t (fmap (either Left (Right . fmap (tmap t))) m))
     
------------------------------------------------
-- NonDet Monad Transformer
------------------------------------------------
      
data NDSig f a = NilT | ConsT a (f a)

instance Functor f => Functor (NDSig f) where
    fmap _ NilT          = NilT
    fmap f (ConsT a fa)  = ConsT (f a) (fmap f fa)

newtype NonDetT m a = L {unL :: m (NDSig (NonDetT m) a)}

runNonDetT  :: NonDetT m a -> m (NDSig (NonDetT m) a)
runNonDetT  = unL

emptyL  :: Monad m => NonDetT m a
emptyL  = L $ return $ NilT

appendL :: Monad m=> NonDetT m a -> NonDetT m a -> NonDetT m a
appendL (L m1) (L m2) = L $ do  l    <- m1
                                case l of
                                     NilT        -> m2
                                     ConsT a l1  -> return (ConsT a (appendL l1 (L m2)))

foldNonDetT :: Monad m => (a -> m b -> m b) -> m b -> NonDetT m a -> m b
foldNonDetT c n (L m) = do  l <- m 
                            case l of 
                             NilT        -> n 
                             ConsT a l1  -> c a (foldNonDetT c n l1)

collectNonDetT     :: Monad m => NonDetT m a -> m [a]
collectNonDetT lt  = foldNonDetT (\a m -> m >>= return . (a:)) (return []) lt

instance MonadT NonDetT where
    lift m            = L $ liftM (`ConsT` emptyL) m
    m `tbind` f       = L $ foldNonDetT   (\a l -> unL $ f a `appendL` L l)
                                          (return NilT)
                                          m

instance FMonadT NonDetT where
    tmap t (L m) = L $ t $ fmap (\lsig  -> case lsig of
                                        NilT       -> NilT
                                        ConsT a l  -> ConsT a (tmap t l)) m

instance Monad m => Monad (NonDetT m) where
    return  = treturn
    (>>=)   = tbind

instance Functor f => Functor (NonDetT f) where
    fmap h (L f) = L $ fmap (fmap h) f 
      
------------------------------------------------
-- Monads
------------------------------------------------

type State s      = StateT s Id

runState s        = runId . runStateT s

type Reader s     = ReaderT s Id

runReader e       = runId . runReaderT e

type Writer w     = WriterT w Id

runWriter         = runId . runWriterT

type Exception x  = ExcT x Id

runException      = runId . runExcT

type Cont r       = ContT r Id

runCont k          = runId . runContT (Id . k)

type NonDet       = NonDetT Id

runNonDet         = runId . runNonDetT 
  
-------------------
-- Identity Monad
-------------------
      
newtype Id a = Id a

runId         :: Id a -> a
runId (Id a)  = a

instance Monad Id where
    return        = Id
    (Id a) >>= f  = f a

instance Functor Id where fmap = liftM
      
-------------------------------------------------
-- Models and Standard Liftings
-------------------------------------------------
      
type ExtModel f g m  = forall a. f (m (g a)) -> m a
type Model f m       = forall a. f (m a) -> m a
type AlgModel f m    = forall a. f a -> m a

toAlg       :: (Functor f, Monad m) => Model f m -> AlgModel f (Cod m)
toAlg op t  = Cod $ \k ->  op (fmap k t)

liftModel     :: (Functor f, Monad m, Functor m, FMonadT t, Monad (t (Cod m))) => 
                 Model f m -> Model f (t m)
liftModel op  = tmap down . join . lift . toAlg op . fmap (tmap lift)

liftAlgModel     :: (MonadT t, Monad m, Functor f) => AlgModel f m -> AlgModel f (t m)
liftAlgModel op  = lift . op

liftExtModel     ::  (  Functor f, Functor g, Monad m, Functor m, 
                        MMonadT t, Functor (t f), Functor (t m)) => 
                     ExtModel f g m -> ExtModel f g (t m)
liftExtModel op  =    tmap (op . fmap deComp . deComp) . 
                      monoidalT . flift . fmap  (monoidalT . fmap flift) 
      
----------------------
-- State Operations
----------------------
      
data StateOp s a = Get (s -> a) | Put s a

instance Functor (StateOp s) where
    fmap f (Get g)    = Get (f . g)
    fmap f (Put s a)  = Put s (f a)

modelStateT            :: Monad m => AlgModel (StateOp s) (StateT s m)
modelStateT (Get g)    = S (\s -> return (g s, s))
modelStateT (Put s a)  = S (\_ -> return (a, s))

getX     :: Monad m => AlgModel (StateOp s) m -> m s
getX op  = op $ Get id

putX       :: Monad m => AlgModel (StateOp s) m -> s -> m ()
putX op s  = op $ Put s ()
      
----------------------
-- Reader Operations
----------------------
      
data ReaderOp s a = Ask (s -> a) | InEnv s a

instance Functor (ReaderOp s) where
    fmap f (Ask g)      = Ask (f . g)
    fmap f (InEnv s a)  = InEnv s (f a)

modelReaderT              :: Monad m => Model (ReaderOp s) (ReaderT s m)
modelReaderT (Ask g)      = R (\s -> runReaderT s (g s))
modelReaderT (InEnv s a)  = R (\_ -> runReaderT s a)

askX     :: Monad m => Model (ReaderOp s) m -> m s
askX op  = op $ Ask return

inEnvX         :: Monad m => Model (ReaderOp s) m -> s -> m a -> m a
inEnvX op s m  = op $ InEnv s m 
      
------------------------
-- Exception Operations
------------------------
      
data ThrowOp x a   = Throw x
data HandleOp x a  = Handle a (x -> a)

instance Functor (ThrowOp x) where
    fmap _ (Throw x) = Throw x

instance Functor (HandleOp x) where
    fmap f (Handle a h) = Handle (f a) (f . h)

modelThrowExcT            :: Monad m => AlgModel (ThrowOp x) (ExcT x m)
modelThrowExcT (Throw x)  = X (return (Left x))

modelHandleExcT               :: Monad m => Model (HandleOp x) (ExcT x m)
modelHandleExcT (Handle m h)  = X (unX m >>= \exa -> case  exa of
                                                Left x  -> unX (h x)
                                                Right a -> return (Right a))

throwX       :: Monad m => AlgModel (ThrowOp x) m -> x -> m a
throwX op x  = op $ Throw x

handleX         :: Monad m => Model(HandleOp x) m -> m a -> (x -> m a) -> m a
handleX op m h  = op $ Handle m h
      
------------------------
-- Writer Operations
------------------------
      
data WriterOp w a = Trace w a

instance Functor (WriterOp w) where
    fmap f (Trace w a) = Trace w (f a)

modelWriterT :: (Monad m, Monoid w) => AlgModel (WriterOp w) (WriterT w m)
modelWriterT (Trace w a)  = W (return (a,w))

traceX       :: (Monad m) => AlgModel (WriterOp w) m -> w -> m ()
traceX op w  = op $ Trace w ()
      
--------------------------
-- Continuation Operations
--------------------------
      
data ContOp r a = Abort r | CallCC ((a -> r) -> a)

instance Functor (ContOp r) where
    fmap _ (Abort r)      = Abort r
    fmap f (CallCC k)     = CallCC (\c -> f (k (c . f)))

modelContT             :: Monad m => AlgModel (ContOp (m r)) (ContT r m)
modelContT (Abort mr)  = C $ \_ -> mr
modelContT (CallCC k)  = C $ \c -> c (k c)

abortX       :: Monad m => AlgModel (ContOp r) m -> r -> m a
abortX op r  = op (Abort r)

callCCX       :: Monad m => AlgModel (ContOp r) m -> ((a -> r) -> a) -> m a
callCCX op f  = op (CallCC f)

callccX       :: Monad m => AlgModel (ContOp r) m -> ((a -> m b) -> m a) -> m a
callccX op f  =  join $ callCCX op (\k -> f (\x -> abortX op (k (return x))))  
      
--------------------------
-- Step Operations
--------------------------
      
newtype StepOp f x = StepOp (f x)

instance (Functor f) => Functor (StepOp f) where 
    fmap h (StepOp fa) = StepOp (fmap h fa)

modelStepT              :: (Functor f, Monad m) => Model (StepOp f) (StepT f m)
modelStepT (StepOp fa)  = T (return (Right fa))

stepX     :: (Monad m) => Model (StepOp f) m -> f (m x) -> m x
stepX op  = op . StepOp 
  
--------------------------
-- NonDet Operations
--------------------------
      
data NonDetOp a = ZeroND | PlusND a a

instance Functor NonDetOp where
    fmap _ ZeroND        = ZeroND
    fmap f (PlusND a b)  = PlusND (f a) (f b)

modelNonDetT               :: Monad m => AlgModel NonDetOp (NonDetT m)
modelNonDetT ZeroND        = emptyL
modelNonDetT (PlusND t u)  = appendL (return t) (return u)

zeroNDX         :: Monad m => AlgModel NonDetOp m -> m a
zeroNDX op      = op ZeroND

plusNDX         :: Monad m => AlgModel NonDetOp m -> m a -> m a -> m a
plusNDX op t u  = join $ op (PlusND t u)
      
  
-----------------------------------------
-- Functor Composition
----------------------------------------
      
newtype (FComp f g) a = Comp {deComp :: (f (g a)) }

instance (Functor f, Functor g) => Functor (FComp f g) where
    fmap f (Comp fga) = Comp (fmap (fmap f) fga)
      
-----------------------------------------
-- Either Functor Instance
----------------------------------------
      
instance Functor (Either x) where
    fmap _ (Left x)   = Left x
    fmap f (Right y)  = Right (f y)

